package org.eclipse.jdt.internal.compiler.util;

import org.eclipse.jdt.internal.compiler.batch.Main;
import java.util.ArrayList;
import java.util.Iterator;
import org.eclipse.jdt.internal.compiler.batch.FileSystem;
import java.util.List;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import java.util.HashSet;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import java.io.BufferedOutputStream;
import org.eclipse.jdt.internal.compiler.ClassFile;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import java.io.ByteArrayInputStream;
import java.util.StringTokenizer;
import java.io.IOException;
import java.io.File;
import org.eclipse.jdt.core.compiler.CharOperation;

public class Util implements SuffixConstants
{
    public static final char C_BOOLEAN = 'Z';
    public static final char C_BYTE = 'B';
    public static final char C_CHAR = 'C';
    public static final char C_DOUBLE = 'D';
    public static final char C_FLOAT = 'F';
    public static final char C_INT = 'I';
    public static final char C_SEMICOLON = ';';
    public static final char C_COLON = ':';
    public static final char C_LONG = 'J';
    public static final char C_SHORT = 'S';
    public static final char C_VOID = 'V';
    public static final char C_TYPE_VARIABLE = 'T';
    public static final char C_STAR = '*';
    public static final char C_EXCEPTION_START = '^';
    public static final char C_EXTENDS = '+';
    public static final char C_SUPER = '-';
    public static final char C_DOT = '.';
    public static final char C_DOLLAR = '$';
    public static final char C_ARRAY = '[';
    public static final char C_RESOLVED = 'L';
    public static final char C_UNRESOLVED = 'Q';
    public static final char C_NAME_END = ';';
    public static final char C_PARAM_START = '(';
    public static final char C_PARAM_END = ')';
    public static final char C_GENERIC_START = '<';
    public static final char C_GENERIC_END = '>';
    public static final char C_CAPTURE = '!';
    private static final int DEFAULT_READING_SIZE = 8192;
    private static final int DEFAULT_WRITING_SIZE = 1024;
    public static final String UTF_8 = "UTF-8";
    public static final String LINE_SEPARATOR;
    public static final String EMPTY_STRING;
    public static final int[] EMPTY_INT_ARRAY;
    
    static {
        LINE_SEPARATOR = System.getProperty("line.separator");
        EMPTY_STRING = new String(CharOperation.NO_CHAR);
        EMPTY_INT_ARRAY = new int[0];
    }
    
    public Util() {
    }
    
    public static String buildAllDirectoriesInto(String outputPath, String relativeFileName) throws IOException {
        char fileSeparatorChar = File.separatorChar;
        String fileSeparator = File.separator;
        outputPath = outputPath.replace('/', fileSeparatorChar);
        relativeFileName = relativeFileName.replace('/', fileSeparatorChar);
        int separatorIndex = relativeFileName.lastIndexOf((int)fileSeparatorChar);
        String outputDirPath;
        String fileName;
        if (separatorIndex == -1) {
            if (outputPath.endsWith(fileSeparator)) {
                outputDirPath = outputPath.substring(0, outputPath.length() - 1);
                fileName = outputPath + relativeFileName;
            }
            else {
                outputDirPath = outputPath;
                fileName = outputPath + fileSeparator + relativeFileName;
            }
        }
        else if (outputPath.endsWith(fileSeparator)) {
            outputDirPath = outputPath + relativeFileName.substring(0, separatorIndex);
            fileName = outputPath + relativeFileName;
        }
        else {
            outputDirPath = outputPath + fileSeparator + relativeFileName.substring(0, separatorIndex);
            fileName = outputPath + fileSeparator + relativeFileName;
        }
        File f = new File(outputDirPath);
        f.mkdirs();
        if (f.isDirectory()) {
            return fileName;
        }
        if (outputPath.endsWith(fileSeparator)) {
            outputPath = outputPath.substring(0, outputPath.length() - 1);
        }
        f = new File(outputPath);
        boolean checkFileType = false;
        if (f.exists()) {
            checkFileType = true;
        }
        else if (!f.mkdirs()) {
            if (f.exists()) {
                checkFileType = true;
            }
            else {
                throw new IOException(Messages.bind(Messages.output_notValidAll, f.getAbsolutePath()));
            }
        }
        if (checkFileType && !f.isDirectory()) {
            throw new IOException(Messages.bind(Messages.output_isFile, f.getAbsolutePath()));
        }
        StringBuffer outDir = new StringBuffer(outputPath);
        outDir.append(fileSeparator);
        StringTokenizer tokenizer = new StringTokenizer(relativeFileName, fileSeparator);
        String token = tokenizer.nextToken();
        while (tokenizer.hasMoreTokens()) {
            f = new File(outDir.append(token).append(fileSeparator).toString());
            checkFileType = false;
            if (f.exists()) {
                checkFileType = true;
            }
            else if (!f.mkdir()) {
                if (f.exists()) {
                    checkFileType = true;
                }
                else {
                    throw new IOException(Messages.bind(Messages.output_notValid, outDir.substring(outputPath.length() + 1, outDir.length() - 1), outputPath));
                }
            }
            if (checkFileType && !f.isDirectory()) {
                throw new IOException(Messages.bind(Messages.output_isFile, f.getAbsolutePath()));
            }
            token = tokenizer.nextToken();
        }
        return outDir.append(token).toString();
    }
    
    public static char[] bytesToChar(byte[] bytes, String encoding) throws IOException {
        return getInputStreamAsCharArray(new ByteArrayInputStream(bytes), bytes.length, encoding);
    }
    
    public static int computeOuterMostVisibility(TypeDeclaration typeDeclaration, int visibility) {
        while (typeDeclaration != null) {
            switch (typeDeclaration.modifiers & 0x7) {
                case 2: {
                    visibility = 2;
                    break;
                }
                case 0: {
                    if (visibility != 2) {
                        visibility = 0;
                        break;
                    }
                    break;
                }
                case 4: {
                    if (visibility == 1) {
                        visibility = 4;
                        break;
                    }
                    break;
                }
            }
            typeDeclaration = typeDeclaration.enclosingType;
        }
        return visibility;
    }
    
    public static byte[] getFileByteContent(File file) throws IOException {
        InputStream stream = null;
        try {
            stream = new BufferedInputStream(new FileInputStream(file));
            return getInputStreamAsByteArray(stream, (int)file.length());
        }
        finally {
            if (stream != null) {
                try {
                    stream.close();
                }
                catch (IOException ex) {}
            }
        }
    }
    
    public static char[] getFileCharContent(File file, String encoding) throws IOException {
        InputStream stream = null;
        try {
            stream = new FileInputStream(file);
            return getInputStreamAsCharArray(stream, (int)file.length(), encoding);
        }
        finally {
            if (stream != null) {
                try {
                    stream.close();
                }
                catch (IOException ex) {}
            }
        }
    }
    
    private static FileOutputStream getFileOutputStream(boolean generatePackagesStructure, String outputPath, String relativeFileName) throws IOException {
        if (generatePackagesStructure) {
            return new FileOutputStream(new File(buildAllDirectoriesInto(outputPath, relativeFileName)));
        }
        String fileName = null;
        char fileSeparatorChar = File.separatorChar;
        String fileSeparator = File.separator;
        outputPath = outputPath.replace('/', fileSeparatorChar);
        int indexOfPackageSeparator = relativeFileName.lastIndexOf((int)fileSeparatorChar);
        if (indexOfPackageSeparator == -1) {
            if (outputPath.endsWith(fileSeparator)) {
                fileName = outputPath + relativeFileName;
            }
            else {
                fileName = outputPath + fileSeparator + relativeFileName;
            }
        }
        else {
            int length = relativeFileName.length();
            if (outputPath.endsWith(fileSeparator)) {
                fileName = outputPath + relativeFileName.substring(indexOfPackageSeparator + 1, length);
            }
            else {
                fileName = outputPath + fileSeparator + relativeFileName.substring(indexOfPackageSeparator + 1, length);
            }
        }
        return new FileOutputStream(new File(fileName));
    }
    
    public static byte[] getInputStreamAsByteArray(InputStream stream, int length) throws IOException {
        byte[] contents;
        if (length == -1) {
            contents = new byte[0];
            int contentsLength = 0;
            int amountRead = -1;
            do {
                int amountRequested = Math.max(stream.available(), 8192);
                if (contentsLength + amountRequested > contents.length) {
                    System.arraycopy(contents, 0, contents = new byte[contentsLength + amountRequested], 0, contentsLength);
                }
                amountRead = stream.read(contents, contentsLength, amountRequested);
                if (amountRead > 0) {
                    contentsLength += amountRead;
                }
                else {
                    continue;
                }
            } while (amountRead != -1);
            if (contentsLength < contents.length) {
                System.arraycopy(contents, 0, contents = new byte[contentsLength], 0, contentsLength);
            }
        }
        else {
            contents = new byte[length];
            for (int len = 0, readSize = 0; readSize != -1 && len != length; len += readSize, readSize = stream.read(contents, len, length - len)) {}
        }
        return contents;
    }
    
    public static char[] getInputStreamAsCharArray(InputStream stream, int length, String encoding) throws IOException {
        BufferedReader reader = null;
        try {
            BufferedReader bufferedReader;
            if (encoding == null) {
                InputStreamReader in2;
                bufferedReader = new BufferedReader(in2);
                in2 = new InputStreamReader(stream);
            }
            else {
                InputStreamReader in;
                bufferedReader = new BufferedReader(in);
                in = new InputStreamReader(stream, encoding);
            }
            reader = bufferedReader;
        }
        catch (UnsupportedEncodingException ex) {
            reader = new BufferedReader(new InputStreamReader(stream));
        }
        int totalRead = 0;
        char[] contents;
        if (length == -1) {
            contents = CharOperation.NO_CHAR;
        }
        else {
            contents = new char[length];
        }
        while (true) {
            int amountRequested;
            if (totalRead < length) {
                amountRequested = length - totalRead;
            }
            else {
                int current = reader.read();
                if (current < 0) {
                    break;
                }
                amountRequested = Math.max(stream.available(), 8192);
                if (totalRead + 1 + amountRequested > contents.length) {
                    System.arraycopy(contents, 0, contents = new char[totalRead + 1 + amountRequested], 0, totalRead);
                }
                contents[totalRead++] = (char)current;
            }
            int amountRead = reader.read(contents, totalRead, amountRequested);
            if (amountRead < 0) {
                break;
            }
            totalRead += amountRead;
        }
        int start = 0;
        if (totalRead > 0 && "UTF-8".equals(encoding) && contents[0] == '\ufeff') {
            --totalRead;
            start = 1;
        }
        if (totalRead < contents.length) {
            System.arraycopy(contents, start, contents = new char[totalRead], 0, totalRead);
            return contents;
        }
        return contents;
    }
    
    public static String getExceptionSummary(Throwable exception) {
        StringWriter stringWriter = new StringWriter();
        exception.printStackTrace(new PrintWriter(stringWriter));
        StringBuffer buffer = stringWriter.getBuffer();
        StringBuffer exceptionBuffer = new StringBuffer(50);
        exceptionBuffer.append(exception.toString());
        int i = 0;
        int lineSep = 0;
        int max = buffer.length();
        int line2Start = 0;
    Label_0169:
        while (i < max) {
            switch (buffer.charAt(i)) {
                case '\n':
                case '\r': {
                    if (line2Start > 0) {
                        exceptionBuffer.append(' ').append(buffer.substring(line2Start, i));
                        break Label_0169;
                    }
                    ++lineSep;
                    break;
                }
                case '\t':
                case ' ': {
                    break;
                }
                default: {
                    if (lineSep > 0) {
                        line2Start = i;
                        lineSep = 0;
                        break;
                    }
                    break;
                }
            }
            ++i;
        }
        return exceptionBuffer.toString();
    }
    
    public static int getLineNumber(int position, int[] lineEnds, int g, int d) {
        if (lineEnds == null) {
            return 1;
        }
        if (d == -1) {
            return 1;
        }
        int m = g;
        while (g <= d) {
            m = g + (d - g) / 2;
            int start;
            if (position < (start = lineEnds[m])) {
                d = m - 1;
            }
            else if (position > start) {
                g = m + 1;
            }
            else {
                return m + 1;
            }
        }
        if (position < lineEnds[m]) {
            return m + 1;
        }
        return m + 2;
    }
    
    public static byte[] getZipEntryByteContent(ZipEntry ze, ZipFile zip) throws IOException {
        InputStream stream = null;
        try {
            InputStream inputStream = zip.getInputStream(ze);
            if (inputStream == null) {
                throw new IOException("Invalid zip entry name : " + ze.getName());
            }
            stream = new BufferedInputStream(inputStream);
            return getInputStreamAsByteArray(stream, (int)ze.getSize());
        }
        finally {
            if (stream != null) {
                try {
                    stream.close();
                }
                catch (IOException ex) {}
            }
        }
    }
    
    public static int hashCode(Object[] array) {
        int prime = 31;
        if (array == null) {
            return 0;
        }
        int result = 1;
        for (int index = 0; index < array.length; ++index) {
            result = prime * result + ((array[index] == null) ? 0 : array[index].hashCode());
        }
        return result;
    }
    
    public static final boolean isPotentialZipArchive(String name) {
        int lastDot = name.lastIndexOf(46);
        if (lastDot == -1) {
            return false;
        }
        if (name.lastIndexOf((int)File.separatorChar) > lastDot) {
            return false;
        }
        int length = name.length();
        int extensionLength = length - lastDot - 1;
        if (extensionLength == "java".length()) {
            int i = extensionLength - 1;
            while (i >= 0) {
                if (Character.toLowerCase(name.charAt(length - extensionLength + i)) != "java".charAt(i)) {
                    break;
                }
                if (i == 0) {
                    return false;
                }
                --i;
            }
        }
        if (extensionLength == "class".length()) {
            int i = extensionLength - 1;
            while (i >= 0) {
                if (Character.toLowerCase(name.charAt(length - extensionLength + i)) != "class".charAt(i)) {
                    return true;
                }
                --i;
            }
            return false;
        }
        return true;
    }
    
    public static final boolean isClassFileName(char[] name) {
        int nameLength = (name == null) ? 0 : name.length;
        int suffixLength = Util.SUFFIX_CLASS.length;
        if (nameLength < suffixLength) {
            return false;
        }
        int i = 0;
        int offset = nameLength - suffixLength;
        while (i < suffixLength) {
            char c = name[offset + i];
            if (c != Util.SUFFIX_class[i] && c != Util.SUFFIX_CLASS[i]) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    public static final boolean isClassFileName(String name) {
        int nameLength = (name == null) ? 0 : name.length();
        int suffixLength = Util.SUFFIX_CLASS.length;
        if (nameLength < suffixLength) {
            return false;
        }
        int i = 0;
        while (i < suffixLength) {
            char c = name.charAt(nameLength - i - 1);
            int suffixIndex = suffixLength - i - 1;
            if (c != Util.SUFFIX_class[suffixIndex] && c != Util.SUFFIX_CLASS[suffixIndex]) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    public static final boolean isExcluded(char[] path, char[][] inclusionPatterns, char[][] exclusionPatterns, boolean isFolderPath) {
        if (inclusionPatterns == null && exclusionPatterns == null) {
            return false;
        }
        Label_0139: {
            if (inclusionPatterns != null) {
                int i = 0;
                int length = inclusionPatterns.length;
                while (i < length) {
                    char[] folderPattern;
                    char[] pattern = folderPattern = inclusionPatterns[i];
                    if (isFolderPath) {
                        int lastSlash = CharOperation.lastIndexOf('/', pattern);
                        if (lastSlash != -1 && lastSlash != pattern.length - 1) {
                            int star = CharOperation.indexOf('*', pattern, lastSlash);
                            if (star == -1 || star >= pattern.length - 1 || pattern[star + 1] != '*') {
                                folderPattern = CharOperation.subarray(pattern, 0, lastSlash);
                            }
                        }
                    }
                    if (CharOperation.pathMatch(folderPattern, path, true, '/')) {
                        break Label_0139;
                    }
                    ++i;
                }
                return true;
            }
        }
        if (isFolderPath) {
            path = CharOperation.concat(path, new char[] { '*' }, '/');
        }
        if (exclusionPatterns != null) {
            int i = 0;
            int length = exclusionPatterns.length;
            while (i < length) {
                if (CharOperation.pathMatch(exclusionPatterns[i], path, true, '/')) {
                    return true;
                }
                ++i;
            }
        }
        return false;
    }
    
    public static final boolean isJavaFileName(char[] name) {
        int nameLength = (name == null) ? 0 : name.length;
        int suffixLength = Util.SUFFIX_JAVA.length;
        if (nameLength < suffixLength) {
            return false;
        }
        int i = 0;
        int offset = nameLength - suffixLength;
        while (i < suffixLength) {
            char c = name[offset + i];
            if (c != Util.SUFFIX_java[i] && c != Util.SUFFIX_JAVA[i]) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    public static final boolean isJavaFileName(String name) {
        int nameLength = (name == null) ? 0 : name.length();
        int suffixLength = Util.SUFFIX_JAVA.length;
        if (nameLength < suffixLength) {
            return false;
        }
        int i = 0;
        while (i < suffixLength) {
            char c = name.charAt(nameLength - i - 1);
            int suffixIndex = suffixLength - i - 1;
            if (c != Util.SUFFIX_java[suffixIndex] && c != Util.SUFFIX_JAVA[suffixIndex]) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    public static void reverseQuickSort(char[][] list, int left, int right) {
        int original_left = left;
        int original_right = right;
        char[] mid = list[left + (right - left) / 2];
        while (true) {
            if (CharOperation.compareTo(list[left], mid) > 0) {
                ++left;
                continue;
            }
            while (CharOperation.compareTo(mid, list[right]) > 0) {
                --right;
            }
            if (left <= right) {
                char[] tmp = list[left];
                list[left] = list[right];
                list[right] = tmp;
                ++left;
                --right;
            }
            if (left > right) {
                break;
            }
            continue;
        }
        if (original_left < right) {
            reverseQuickSort(list, original_left, right);
        }
        if (left < original_right) {
            reverseQuickSort(list, left, original_right);
        }
    }
    
    public static void reverseQuickSort(char[][] list, int left, int right, int[] result) {
        int original_left = left;
        int original_right = right;
        char[] mid = list[left + (right - left) / 2];
        while (true) {
            if (CharOperation.compareTo(list[left], mid) > 0) {
                ++left;
                continue;
            }
            while (CharOperation.compareTo(mid, list[right]) > 0) {
                --right;
            }
            if (left <= right) {
                char[] tmp = list[left];
                list[left] = list[right];
                list[right] = tmp;
                int temp = result[left];
                result[left] = result[right];
                result[right] = temp;
                ++left;
                --right;
            }
            if (left > right) {
                break;
            }
            continue;
        }
        if (original_left < right) {
            reverseQuickSort(list, original_left, right, result);
        }
        if (left < original_right) {
            reverseQuickSort(list, left, original_right, result);
        }
    }
    
    public static final int searchColumnNumber(int[] startLineIndexes, int lineNumber, int position) {
        switch (lineNumber) {
            case 1: {
                return position + 1;
            }
            case 2: {
                return position - startLineIndexes[0];
            }
            default: {
                int line = lineNumber - 2;
                int length = startLineIndexes.length;
                if (line >= length) {
                    return position - startLineIndexes[length - 1];
                }
                return position - startLineIndexes[line];
            }
        }
    }
    
    public static Boolean toBoolean(boolean bool) {
        if (bool) {
            return Boolean.TRUE;
        }
        return Boolean.FALSE;
    }
    
    public static String toString(Object[] objects) {
        return toString(objects, new Util.Displayable() {
            @Override
            public String displayString(Object o) {
                if (o == null) {
                    return "null";
                }
                return o.toString();
            }
        });
    }
    
    public static String toString(Object[] objects, Util.Displayable renderer) {
        if (objects == null) {
            return "";
        }
        StringBuffer buffer = new StringBuffer(10);
        for (int i = 0; i < objects.length; ++i) {
            if (i > 0) {
                buffer.append(", ");
            }
            buffer.append(renderer.displayString(objects[i]));
        }
        return buffer.toString();
    }
    
    public static void writeToDisk(boolean generatePackagesStructure, String outputPath, String relativeFileName, ClassFile classFile) throws IOException {
        FileOutputStream file = getFileOutputStream(generatePackagesStructure, outputPath, relativeFileName);
        BufferedOutputStream output = new BufferedOutputStream(file, 1024);
        try {
            output.write(classFile.header, 0, classFile.headerOffset);
            output.write(classFile.contents, 0, classFile.contentsOffset);
            output.flush();
        }
        catch (IOException e) {
            throw e;
        }
        finally {
            output.close();
        }
    }
    
    public static void recordNestedType(ClassFile classFile, TypeBinding typeBinding) {
        if (classFile.visitedTypes == null) {
            classFile.visitedTypes = new HashSet(3);
        }
        else if (classFile.visitedTypes.contains(typeBinding)) {
            return;
        }
        classFile.visitedTypes.add(typeBinding);
        if (typeBinding.isParameterizedType() && (typeBinding.tagBits & 0x800L) != 0x0L) {
            ParameterizedTypeBinding parameterizedTypeBinding = (ParameterizedTypeBinding)typeBinding;
            ReferenceBinding genericType = parameterizedTypeBinding.genericType();
            if ((genericType.tagBits & 0x800L) != 0x0L) {
                recordNestedType(classFile, genericType);
            }
            TypeBinding[] arguments = parameterizedTypeBinding.arguments;
            if (arguments != null) {
                for (int j = 0, max4 = arguments.length; j < max4; ++j) {
                    TypeBinding argument = arguments[j];
                    if (argument.isWildcard()) {
                        WildcardBinding wildcardBinding = (WildcardBinding)argument;
                        TypeBinding bound = wildcardBinding.bound;
                        if (bound != null && (bound.tagBits & 0x800L) != 0x0L) {
                            recordNestedType(classFile, bound);
                        }
                        ReferenceBinding superclass = wildcardBinding.superclass();
                        if (superclass != null && (superclass.tagBits & 0x800L) != 0x0L) {
                            recordNestedType(classFile, superclass);
                        }
                        ReferenceBinding[] superInterfaces = wildcardBinding.superInterfaces();
                        if (superInterfaces != null) {
                            for (int i = 0, max5 = superInterfaces.length; i < max5; ++i) {
                                ReferenceBinding superInterface = superInterfaces[i];
                                if ((superInterface.tagBits & 0x800L) != 0x0L) {
                                    recordNestedType(classFile, superInterface);
                                }
                            }
                        }
                    }
                    else if ((argument.tagBits & 0x800L) != 0x0L) {
                        recordNestedType(classFile, argument);
                    }
                }
            }
        }
        else if (typeBinding.isTypeVariable() && (typeBinding.tagBits & 0x800L) != 0x0L) {
            TypeVariableBinding typeVariableBinding = (TypeVariableBinding)typeBinding;
            TypeBinding upperBound = typeVariableBinding.upperBound();
            if (upperBound != null && (upperBound.tagBits & 0x800L) != 0x0L) {
                recordNestedType(classFile, upperBound);
            }
            TypeBinding[] upperBounds = typeVariableBinding.otherUpperBounds();
            if (upperBounds != null) {
                for (int k = 0, max3 = upperBounds.length; k < max3; ++k) {
                    TypeBinding otherUpperBound = upperBounds[k];
                    if ((otherUpperBound.tagBits & 0x800L) != 0x0L) {
                        recordNestedType(classFile, otherUpperBound);
                    }
                }
            }
        }
        else if (typeBinding.isNestedType()) {
            TypeBinding enclosingType = typeBinding;
            Label_0448: {
                while (enclosingType.canBeSeenBy(classFile.referenceBinding.scope)) {
                    enclosingType = enclosingType.enclosingType();
                    if (enclosingType != null) {
                        continue;
                    }
                    boolean onBottomForBug445231 = enclosingType != null;
                    classFile.recordInnerClasses(typeBinding, onBottomForBug445231);
                    return;
                }
                continue;
            }
        }
    }
    
    public static File getJavaHome() {
        String javaHome = System.getProperty("java.home");
        if (javaHome != null) {
            File javaHomeFile = new File(javaHome);
            if (javaHomeFile.exists()) {
                return javaHomeFile;
            }
        }
        return null;
    }
    
    public static void collectRunningVMBootclasspath(List bootclasspaths) {
        for (String filePath : collectFilesNames()) {
            FileSystem.Classpath currentClasspath = FileSystem.getClasspath(filePath, null, null, null);
            if (currentClasspath != null) {
                bootclasspaths.add(currentClasspath);
            }
            else {
                continue;
            }
        }
    }
    
    public static List<String> collectFilesNames() {
        String javaversion = System.getProperty("java.version");
        if (javaversion != null && javaversion.equalsIgnoreCase("1.1.8")) {
            throw new IllegalStateException();
        }
        String bootclasspathProperty = System.getProperty("sun.boot.class.path");
        if (bootclasspathProperty == null || bootclasspathProperty.length() == 0) {
            bootclasspathProperty = System.getProperty("vm.boot.class.path");
            if (bootclasspathProperty == null || bootclasspathProperty.length() == 0) {
                bootclasspathProperty = System.getProperty("org.apache.harmony.boot.class.path");
            }
        }
        List<String> filePaths = new ArrayList();
        if (bootclasspathProperty != null && bootclasspathProperty.length() != 0) {
            StringTokenizer tokenizer = new StringTokenizer(bootclasspathProperty, File.pathSeparator);
            while (tokenizer.hasMoreTokens()) {
                filePaths.add(tokenizer.nextToken());
            }
        }
        else {
            File javaHome = getJavaHome();
            if (javaHome != null) {
                File[] directoriesToCheck = null;
                if (System.getProperty("os.name").startsWith("Mac")) {
                    directoriesToCheck = new File[] { new File(javaHome, "../Classes") };
                }
                else {
                    directoriesToCheck = new File[] { new File(javaHome, "lib") };
                }
                File[][] systemLibrariesJars = Main.getLibrariesFiles(directoriesToCheck);
                if (systemLibrariesJars != null) {
                    for (int i = 0, max = systemLibrariesJars.length; i < max; ++i) {
                        File[] current = systemLibrariesJars[i];
                        if (current != null) {
                            for (int j = 0, max2 = current.length; j < max2; ++j) {
                                filePaths.add(current[j].getAbsolutePath());
                            }
                        }
                    }
                }
            }
        }
        return filePaths;
    }
    
    public static int getParameterCount(char[] methodSignature) {
        try {
            int count = 0;
            int i = CharOperation.indexOf('(', methodSignature);
            if (i < 0) {
                throw new IllegalArgumentException();
            }
            ++i;
            while (methodSignature[i] != ')') {
                int e = scanTypeSignature(methodSignature, i);
                if (e < 0) {
                    throw new IllegalArgumentException();
                }
                i = e + 1;
                ++count;
            }
            return count;
        }
        catch (ArrayIndexOutOfBoundsException ex) {
            throw new IllegalArgumentException();
        }
    }
    
    public static int scanTypeSignature(char[] string, int start) {
        if (start >= string.length) {
            throw new IllegalArgumentException();
        }
        char c = string[start];
        switch (c) {
            case '[': {
                return scanArrayTypeSignature(string, start);
            }
            case 'L':
            case 'Q': {
                return scanClassTypeSignature(string, start);
            }
            case 'T': {
                return scanTypeVariableSignature(string, start);
            }
            case 'B':
            case 'C':
            case 'D':
            case 'F':
            case 'I':
            case 'J':
            case 'S':
            case 'V':
            case 'Z': {
                return scanBaseTypeSignature(string, start);
            }
            case '!': {
                return scanCaptureTypeSignature(string, start);
            }
            case '*':
            case '+':
            case '-': {
                return scanTypeBoundSignature(string, start);
            }
            default: {
                throw new IllegalArgumentException();
            }
        }
    }
    
    public static int scanBaseTypeSignature(char[] string, int start) {
        if (start >= string.length) {
            throw new IllegalArgumentException();
        }
        char c = string[start];
        if ("BCDFIJSVZ".indexOf((int)c) >= 0) {
            return start;
        }
        throw new IllegalArgumentException();
    }
    
    public static int scanArrayTypeSignature(char[] string, int start) {
        int length = string.length;
        if (start >= length - 1) {
            throw new IllegalArgumentException();
        }
        char c = string[start];
        if (c != '[') {
            throw new IllegalArgumentException();
        }
        c = string[++start];
        while (c == '[') {
            if (start >= length - 1) {
                throw new IllegalArgumentException();
            }
            c = string[++start];
        }
        return scanTypeSignature(string, start);
    }
    
    public static int scanCaptureTypeSignature(char[] string, int start) {
        if (start >= string.length - 1) {
            throw new IllegalArgumentException();
        }
        char c = string[start];
        if (c != '!') {
            throw new IllegalArgumentException();
        }
        return scanTypeBoundSignature(string, start + 1);
    }
    
    public static int scanTypeVariableSignature(char[] string, int start) {
        if (start >= string.length - 2) {
            throw new IllegalArgumentException();
        }
        char c = string[start];
        if (c != 'T') {
            throw new IllegalArgumentException();
        }
        int id = scanIdentifier(string, start + 1);
        c = string[id + 1];
        if (c == ';') {
            return id + 1;
        }
        throw new IllegalArgumentException();
    }
    
    public static int scanIdentifier(char[] string, int start) {
        if (start >= string.length) {
            throw new IllegalArgumentException();
        }
        int p = start;
        do {
            char c = string[p];
            if (c == '<' || c == '>' || c == ':' || c == ';' || c == '.' || c == '/') {
                return p - 1;
            }
            continue;
        } while (++p != string.length);
        return p - 1;
    }
    
    public static int scanClassTypeSignature(char[] string, int start) {
        if (start >= string.length - 2) {
            throw new IllegalArgumentException();
        }
        char c = string[start];
        if (c != 'L' && c != 'Q') {
            return -1;
        }
        int p = start + 1;
        while (p < string.length) {
            c = string[p];
            if (c == ';') {
                return p;
            }
            if (c == '<') {
                int e = p = scanTypeArgumentSignatures(string, p);
            }
            else if (c == '.' || c == '/') {
                int id = p = scanIdentifier(string, p + 1);
            }
            ++p;
        }
        throw new IllegalArgumentException();
    }
    
    public static int scanTypeBoundSignature(char[] string, int start) {
        if (start >= string.length) {
            throw new IllegalArgumentException();
        }
        char c = string[start];
        switch (c) {
            case '*': {
                return start;
            }
            case '+':
            case '-': {
                if (start >= string.length - 2) {
                    throw new IllegalArgumentException();
                }
                c = string[++start];
                switch (c) {
                    case '!': {
                        return scanCaptureTypeSignature(string, start);
                    }
                    case '+':
                    case '-': {
                        return scanTypeBoundSignature(string, start);
                    }
                    case 'L':
                    case 'Q': {
                        return scanClassTypeSignature(string, start);
                    }
                    case 'T': {
                        return scanTypeVariableSignature(string, start);
                    }
                    case '[': {
                        return scanArrayTypeSignature(string, start);
                    }
                    case '*': {
                        return start;
                    }
                    default: {
                        throw new IllegalArgumentException();
                    }
                }
            }
            default: {
                throw new IllegalArgumentException();
            }
        }
    }
    
    public static int scanTypeArgumentSignatures(char[] string, int start) {
        if (start >= string.length - 1) {
            throw new IllegalArgumentException();
        }
        char c = string[start];
        if (c != '<') {
            throw new IllegalArgumentException();
        }
        int p = start + 1;
        while (p < string.length) {
            c = string[p];
            if (c == '>') {
                return p;
            }
            int e = scanTypeArgumentSignature(string, p);
            p = e + 1;
        }
        throw new IllegalArgumentException();
    }
    
    public static int scanTypeArgumentSignature(char[] string, int start) {
        if (start >= string.length) {
            throw new IllegalArgumentException();
        }
        char c = string[start];
        switch (c) {
            case '*': {
                return start;
            }
            case '+':
            case '-': {
                return scanTypeBoundSignature(string, start);
            }
            default: {
                return scanTypeSignature(string, start);
            }
        }
    }
    
    public static boolean effectivelyEqual(Object[] one, Object[] two) {
        if (one == two) {
            return true;
        }
        int oneLength = (one == null) ? 0 : one.length;
        int twoLength = (two == null) ? 0 : two.length;
        if (oneLength != twoLength) {
            return false;
        }
        if (oneLength == 0) {
            return true;
        }
        int i = 0;
        while (i < one.length) {
            if (one[i] != two[i]) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    public static void appendEscapedChar(StringBuffer buffer, char c, boolean stringLiteral) {
        switch (c) {
            case '\b': {
                buffer.append("\\b");
                break;
            }
            case '\t': {
                buffer.append("\\t");
                break;
            }
            case '\n': {
                buffer.append("\\n");
                break;
            }
            case '\f': {
                buffer.append("\\f");
                break;
            }
            case '\r': {
                buffer.append("\\r");
                break;
            }
            case '\"': {
                if (stringLiteral) {
                    buffer.append("\\\"");
                    break;
                }
                buffer.append(c);
                break;
            }
            case '\'': {
                if (stringLiteral) {
                    buffer.append(c);
                    break;
                }
                buffer.append("\\'");
                break;
            }
            case '\\': {
                buffer.append("\\\\");
                break;
            }
            default: {
                if (c >= ' ') {
                    buffer.append(c);
                    break;
                }
                if (c >= '\u0010') {
                    buffer.append("\\u00").append(Integer.toHexString((int)c));
                    break;
                }
                if (c >= '\0') {
                    buffer.append("\\u000").append(Integer.toHexString((int)c));
                    break;
                }
                buffer.append(c);
                break;
            }
        }
    }
    
    public interface Displayable
    {
        public abstract String displayString(Object p0);
    }
}
